"""
Abstract interfaces for root cause analysis
"""
from abc import ABC
from typing import Callable
from typing import Dict
from typing import Sequence
from typing import Union

from ..model.case import CaseData
from ..model.graph import Graph
from ..model.graph import Node


class Score:
    """
    The more suspicious a node, the higher the score.
    """

    def __init__(self, score: float, info: dict = None, key: tuple = None):
        self._score = score
        self._key = key
        self._info = {} if info is None else info

    def __eq__(self, obj) -> bool:
        if isinstance(obj, Score):
            return self.score == obj.score
        return False

    def __getitem__(self, key: str):
        return self._info[key]

    def __setitem__(self, key: str, value):
        self._info[key] = value

    def get(self, key: str, default=None):
        """
        Return the value for key if key has been set, else default
        """
        return self._info.get(key, default)

    @property
    def score(self) -> float:
        """
        The overall score
        """
        return self._score

    @score.setter
    def score(self, value: float):
        """
        Update score
        """
        self._score = value

    @property
    def key(self) -> float:
        """
        key for sorting
        """
        return self.score if self._key is None else self._key

    @key.setter
    def key(self, value: tuple):
        """
        Update key
        """
        self._key = value

    @property
    def info(self) -> dict:
        """
        Extra information
        """
        return self._info

    def update(self, score: "Score") -> "Score":
        """
        Update score and info
        """
        self._info.update(score.info)
        self.score = score.score
        self.key = score.key
        return self

    def asdict(self) -> Dict[str, Union[float, dict, tuple]]:
        """
        Serialized as a dict
        """
        data = {"score": self._score, "info": {**self._info}}
        if self._key is not None:
            data["key"] = self._key
        return data

    def __repr__(self) -> str:
        return str(self.asdict())


class Scorer(ABC):
    """
    The abstract interface to score nodes
    """

    def __init__(
        self,
        aggregator: Callable[[Sequence[float]], float] = max,
        max_workers: int = 1,
        seed: int = 0,
        cuda: bool = False,
    ):
        self._aggregator = aggregator
        self._max_workers = max_workers
        self._seed = seed
        self._cuda = cuda

    def score(
        self,
        graph: Graph,
        data: CaseData,
        current: float,
        scores: Dict[Node, Score] = None,
    ) -> Dict[Node, Score]:
        """
        Estimate suspicious nodes
        """
        raise NotImplementedError



import os

import setuptools


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(BASE_DIR, "README.md"), "r", encoding="UTF-8") as readme:
    long_description = readme.read()
with open(
    os.path.join(BASE_DIR, "build", "requirements.txt"), "r", encoding="UTF-8"
) as requirements:
    install_requires = requirements.read().split("\n")


setuptools.setup(
    name="circa-clue",
    version="0.1.1",
    author="limjcst",
    author_email="limjcst@163.com",
    description="Causal Inference-based Root Cause Analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NetManAIOps/CIRCA",
    packages=setuptools.find_packages(include=["circa", "circa.*"]),
    package_data={
        "circa.graph.r": ["*.R"],
    },
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
    ],
    license="BSD 3-Clause License",
)



"""
Depth-first searching alone anomalies
"""
from typing import Dict
from typing import Sequence
from typing import Set

import numpy as np

from .base import Score
from .base import Scorer
from .common import pearson
from ..model.case import CaseData
from ..model.graph import Graph
from ..model.graph import Node


class DFSScorer(Scorer):
    """
    Filter in anomalous nodes that do not have anomalous parents in the graph,
    through searching from the SLI.
    """

    def __init__(self, anomaly_threshold: float, **kwargs):
        super().__init__(**kwargs)
        self._anomaly_threshold = anomaly_threshold

    def _extended(
        self, node: Node, parent: Node, anomalies: Set[Node], data: CaseData
    ) -> bool:
        # pylint: disable=unused-argument, no-self-use
        return parent in anomalies

    def score(
        self,
        graph: Graph,
        data: CaseData,
        current: float,
        scores: Dict[Node, Score] = None,
    ) -> Dict[Node, Score]:
        anomalies = {
            node
            for node, score in scores.items()
            if score.score >= self._anomaly_threshold
        }
        layer = {data.sli}
        visited = {data.sli}
        roots: Set[Node] = set()
        while layer:
            next_layer: Set[Node] = set()
            for node in layer:
                is_root = True
                for parent in graph.parents(node):
                    if parent in visited:
                        is_root = False
                    elif self._extended(node, parent, anomalies, data):
                        is_root = False
                        visited.add(parent)
                        next_layer.add(parent)
                if is_root:
                    roots.add(node)
            layer = next_layer

        return {node: scores[node] for node in roots if node in scores}


class MicroHECLScorer(DFSScorer):
    """
    Extend DFSScorer, stopping if the correlation between two linked nodes
    is lower than the given threshold.

    See MicroHECL in ICSE-SEIP'21
    https://doi.org/10.1109/ICSE-SEIP52600.2021.00043

    ATTENTION: This implementation is not thread-safe
    """

    def __init__(self, stop_threshold: float = 0.7, **kwargs):
        super().__init__(**kwargs)
        self._stop_threshold = stop_threshold
        self._data: Dict[Node, Sequence[float]] = None

    def _extended(
        self, node: Node, parent: Node, anomalies: Set[Node], data: CaseData
    ) -> bool:
        if parent not in anomalies:
            return False
        series_a = np.array(self._data[node][-data.test_window :])
        series_b = np.array(self._data[parent][-data.test_window :])
        correlation = pearson(series_a, series_b)
        return correlation >= self._stop_threshold

    def score(
        self,
        graph: Graph,
        data: CaseData,
        current: float,
        scores: Dict[Node, Score] = None,
    ) -> Dict[Node, Score]:
        self._data = data.load_data(graph, current)
        scores = super().score(graph=graph, data=data, current=current, scores=scores)
        self._data = None
        return scores



"""
Cluster Ranking based fault Diagnosis

See ICDM'17
https://doi.org/10.1109/ICDM.2017.129
"""
import numpy as np
import torch
from torch.optim import SGD

from .enmf import ENMF
from .enmf import InvariantNetwork


class CRD(ENMF):
    # pylint: disable=invalid-name
    """
    Cluster Ranking based fault Diagnosis
    """

    def __init__(
        self,
        num_cluster: int = 2,
        alpha: float = 1.2,
        beta: float = 0.1,
        lambda1: float = 0.1,
        gamma: float = 0.5,
        tau: float = 1.0,
        epoches: int = 3000,
        learning_rate: float = 0.1,
        invariant_network: InvariantNetwork = None,
        use_sgd: bool = True,
        cuda: bool = False,
        seed: int = None,
    ):
        # pylint: disable=too-many-arguments
        """
        Constructor

        Parameters:
            num_cluster: The number of clusters in an invariant network
            alpha: A parameter in the Dirichlet distribution with alpha >= 1
            beta: A parameter to balance the object functions for network clustering
                and broken score
            lambda1: l1 penalty parameter
            gamma: A parameter to balance
                the award for neighboring nodes to have similar status scores, and
                the penalty of large bias from the initial seeds
            tau: A larger tau typically results in more zeros in E
        """
        super().__init__(
            gamma=gamma, tau=tau, epoches=epoches, invariant_network=invariant_network
        )
        if seed is not None:
            torch.manual_seed(seed)
            if cuda:
                torch.cuda.manual_seed(seed)

        # Parameters for broken cluster identification
        self._num_cluster = num_cluster
        self._alpha = alpha
        self._beta = beta
        self._lambda1 = lambda1

        self._learning_rate = learning_rate
        self._use_sgd = use_sgd
        self._cuda = cuda

    @property
    def _device(self):
        return "cuda" if self._cuda else "cpu"

    def _cluster_broken(self, A: torch.FloatTensor, B: torch.FloatTensor):
        """
        Broken Cluster Identification

        Parameters:
            A: invariant_matrix
            B: broken_matrix
        Returns:
            U: Cluster membership matrix with the shape of (num_nodes, num_cluster)
            s: Broken score vector with the shape of (num_nodes, )
        """
        # TODO: Get rid of torch
        # As the authors state that, they "take an alternating minimization framework
        # that alternately solves U and s until a stationary point is achieved" to
        # identify broken clusters.

        # The indicator matrix, with W_{xy} = 1 iff (x, y) is invariant but not broken
        W = (A > 0).float() - (B > 0).float()
        U_origin = torch.randn(
            A.size()[0], self._num_cluster, requires_grad=True, device=self._device
        )
        s_origin = torch.rand(
            self._num_cluster, requires_grad=True, device=self._device
        )
        optimizer = SGD([U_origin, s_origin], lr=self._learning_rate, momentum=0.1)

        for epoch in range(self._epoches):
            optimizer.zero_grad()
            U, s = torch.softmax(U_origin, dim=1), torch.sigmoid(s_origin)

            # Eq.(4)
            A_hat = (U / torch.sum(U, dim=0, keepdim=True)) @ U.transpose(0, 1)
            # Eq.(6)
            J_A = -torch.sum(A * torch.log(A_hat)) - (self._alpha - 1) * torch.sum(
                torch.log(U)
            )
            P = (U * s.view(1, -1)) @ U.transpose(0, 1)  # Eq.(7)
            # Eq.(10)
            J_B = -torch.sum(B * torch.log(P)) - torch.sum(W * torch.log(1 - P))

            J_CR: torch.FloatTensor = (
                J_A + self._beta * J_B + self._lambda1 * U_origin.norm(p=1)
            )
            J_CR.backward()
            optimizer.step()

            self._logger.debug(
                "Epoch of Broken Cluster Identification: %04d, "
                "J_A: %06f, J_B: %06f, J_CR: %06f",
                epoch,
                J_A.item(),
                J_B.item(),
                J_CR.item(),
            )
        with torch.no_grad():
            return torch.softmax(U_origin, dim=1), torch.sigmoid(s_origin)

    def _rank_anomaly_sgd(
        self,
        A: torch.FloatTensor,
        B: torch.FloatTensor,
        U: torch.FloatTensor,
    ):
        """
        Causal Anomaly Ranking with Stochastic Gradient Descent
        """
        H = torch.FloatTensor(self.degree_normalize(A.cpu().numpy()))
        if self._cuda:
            H = H.cuda()
        C = (A > 0).float()

        E_origin = torch.randn(
            A.size()[0], self._num_cluster, requires_grad=True, device=self._device
        )
        optimizer = SGD([E_origin], lr=self._learning_rate)

        for epoch in range(self._epoches):
            optimizer.zero_grad()
            E = torch.sigmoid(E_origin)

            status_scores = H @ (U * E)
            # Square of Frobenius norm
            reconstruct_loss: torch.FloatTensor = (
                C * (status_scores @ status_scores.transpose(0, 1)) - B
            ).norm(p="fro") ** 2
            J_H: torch.FloatTensor = reconstruct_loss + self._tau * E.norm(p=1)
            J_H.backward()
            optimizer.step()

            self._logger.debug(
                "Epoch of Causal Anomaly Ranking: %04d, "
                "J_H: %06f, reconstruct_loss: %06f",
                epoch,
                J_H.item(),
                reconstruct_loss.item(),
            )
        with torch.no_grad():
            return torch.sigmoid(E_origin)

    def _rank_anomaly(
        self,
        A: torch.FloatTensor,
        B: torch.FloatTensor,
        U: torch.FloatTensor,
    ):
        """
        Causal Anomaly Ranking
        """
        if self._use_sgd:
            return self._rank_anomaly_sgd(A=A, B=B, U=U)
        E = super().enmf(
            invariant=A.numpy(),
            broken=B.numpy(),
            causes=np.ones((A.shape[0], self._num_cluster)),
            weight=U.cpu().numpy(),
        )
        return torch.FloatTensor(E, device=U.device)

    def rank(self, data: np.ndarray, discrete: bool = True) -> np.ndarray:
        A = self._invariant_network.invariant_matrix(discrete=discrete)
        if np.all(A == 0):
            return np.zeros(A.shape[0])
        B = self._invariant_network.broken_matrix(data=data, discrete=discrete)
        A, B = torch.FloatTensor(A), torch.FloatTensor(B)
        if self._cuda:
            A, B = A.cuda(), B.cuda()

        U, s = self._cluster_broken(A=A, B=B)
        E = self._rank_anomaly(A=A, B=B, U=U)
        scores = (U * E) @ s

        return scores.cpu().numpy()



"""
Abstract interfaces to regress one node on its parents
"""
from abc import ABC
import logging

import numpy as np

from ..common import zscore


class Regressor(ABC):
    """
    Regress one node on its parents, assuming x ~ P(x | pa(X))
    """

    def __init__(self):
        klass = self.__class__
        self._logger = logging.getLogger(f"{klass.__module__}.{klass.__name__}")

    @staticmethod
    def _zscore(train_y: np.ndarray, test_y: np.ndarray) -> np.ndarray:
        return zscore(train_y=train_y, test_y=test_y)

    def _score(
        self,
        train_x: np.ndarray,
        test_x: np.ndarray,
        train_y: np.ndarray,
        test_y: np.ndarray,
    ) -> np.ndarray:
        raise NotImplementedError

    def score(
        self,
        train_x: np.ndarray,
        test_x: np.ndarray,
        train_y: np.ndarray,
        test_y: np.ndarray,
    ) -> np.ndarray:
        """
        Estimate to what extend each value in test_y violates regression
        """
        if len(train_x) == 0:
            return self._zscore(train_y=train_y, test_y=test_y)
        try:
            return self._score(
                train_x=train_x, test_x=test_x, train_y=train_y, test_y=test_y
            )
        except ValueError as err:
            self._logger.warning(err, exc_info=True)
            return self._zscore(train_y=train_y, test_y=test_y)
